"""Principal classes."""
import json
import re
from typing import Dict, List, Optional

from pydantic import BaseModel

from .effective_arp import EffectiveARP


class PrincipalType(str):
    """A principal type, e.g. Federated or AWS.

    See `AWS JSON policy elements: Principal
    <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_principal.html>`__
    for more.
    """


class PrincipalValue(str):
    """An ARN, wildcard, or other appropriate value of a policy Principal.

    See `AWS JSON policy elements: Principal
    <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_principal.html>`__
    for more.
    """


class PrincipalCollection(Dict[PrincipalType, List[PrincipalValue]]):
    """A collection of Principals of different types, unique to PolicyGlass."""

    @property
    def principals(self) -> List["Principal"]:
        result = []
        for principal_type, principal_values in self.items():
            for principal_value in principal_values:
                result.append(Principal(principal_type, principal_value))
        return result

    def __hash__(self) -> int:  # type: ignore[override]
        """Generate a hash for this principal."""
        return hash(json.dumps({"candidate": 5, "data": 1}, sort_keys=True))

    def __lt__(self, other: object) -> bool:
        """There are few scenarios in which a Principal can be said to contain another object.

        "You cannot use a wildcard to match part of a principal name or ARN."
        `AWS JSON policy elements: Principal
        <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_principal.html>`__

        You can however use *just* a wildcard to match a whole principal.
        An ``arn:aws:iam::123456789012:root`` ARN also matches every principal in that account.

        Parameters:
            other: The object to see if this principal contains.
        """
        return False

    def __contains__(self, other: object) -> bool:
        """Not Implemented.

        Parameters:
            other: The object to see if this object contains.

        Raises:
            NotImplementedError: This method is not implemented.
        """
        raise NotImplementedError()


class Principal(BaseModel):
    """A class which represents a single Principal including its type.

    Objects of this type are typically generated by the :class:`~policyglass.statement.Statement` class.
    """

    #: Principal Type
    type: PrincipalType
    #: Principal value
    value: PrincipalValue

    def __init__(self, type: PrincipalType, value: PrincipalValue) -> None:
        super().__init__(
            type=type,
            value=self._normalize_account_id(value),
        )

    def issubset(self, other: object) -> bool:
        """Whether this object contains all the elements of another object (i.e. is a subset of the other object).

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        if self.type != other.type:
            return False
        if other.value == "*":
            return True
        if other.is_account and self.account_id == other.account_id:
            return True
        for self_element, other_element in zip(self.arn_elements, other.arn_elements):
            if self_element != other_element and other_element:
                return False
        return True

    @property
    def account_id(self) -> Optional[str]:
        """Return the account id of this Principal if there is one."""
        try:
            return self.arn_elements[4]
        except IndexError:
            return None

    @property
    def arn_elements(self) -> List[str]:
        """Return a list of arn elements, replacing blanks with ``""``."""
        return [element or "" for element in self.value.split(":")]

    @property
    def is_account(self) -> bool:
        """Return true if the prinncipal is an account."""
        return bool(self.type == "AWS" and re.match(r"^arn:aws:iam::\d+:root$", self.value))

    @staticmethod
    def _normalize_account_id(value: PrincipalValue) -> PrincipalValue:
        """Return a fully qualified account id if a value is a short account id.

        Parameters:
            value: The value to normalize.
        """
        if re.match(r"^\d+$", value):
            return PrincipalValue(f"arn:aws:iam::{value}:root")
        return PrincipalValue(value)

    def __eq__(self, other: object) -> bool:
        """Whether this object contains (but is not equal to) another object.

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        return self.type == other.type and self.value == other.value

    def __lt__(self, other: object) -> bool:
        """Whether this object contains (but is not equal to) another object.

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        if self == other:
            return False
        return self.issubset(other)

    def __repr__(self) -> str:
        """Return an insantiable representation of this object."""
        return f"{self.__class__.__name__}(type='{self.type}', value='{self.value}')"

    def __hash__(self) -> int:
        """Return a hash representation of this object."""
        return hash(str(self))


class EffectivePrincipal(EffectiveARP[Principal]):
    """EffectivePrincipals are the representation of the difference between an Principal and its exclusion.

    The allowed Principal is the difference (subtraction) of the excluded Principals
    from the included Principal.
    """

    _arp_type = Principal



"""Statement Condition classes."""


from typing import Dict, List

from pydantic import BaseModel

from .models import CaseInsensitiveString


class ConditionKey(CaseInsensitiveString):
    """Condition Keys are case insensitive.

    "Condition key names are not case-sensitive."
    - `IAM Reference Policy Elements
    <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_condition.html>`__
    """

    ...


class ConditionOperator(CaseInsensitiveString):
    """Condition Operator.

    See `IAM JSON policy elements: Condition operators
    <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_condition_operators.html>`__
    for more.
    """


class ConditionValue(str):
    """Condition values may or may not be case sensitive depending on the operator."""


class Condition(BaseModel):
    """A representation of part of a statement condition in order to facilitate comparison."""

    key: ConditionKey
    operator: ConditionOperator
    values: List[ConditionValue]

    def __init__(self, key: ConditionKey, operator: ConditionOperator, values: List[ConditionValue]) -> None:
        super().__init__(
            key=key,
            operator=operator,
            values=values,
        )

    @classmethod
    def factory(cls, condition_collection: "ConditionCollection") -> List["Condition"]:
        result = []
        for key, operator_values in condition_collection.items():
            for operator, values in operator_values.items():
                result.append(
                    cls(ConditionKey(key), ConditionOperator(operator), [ConditionValue(value) for value in values])
                )
        return result

    def __eq__(self, other: object) -> bool:
        """Determine whether this object and another object are equal.

        Parameters:
            other: The object to compare this one to.

        Raises:
            ValueError: When the object we are compared with is not of the same type.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        return self.key == other.key and self.operator == other.operator and self.values == other.values

    def __repr__(self) -> str:
        """Return an instantiable representation of the object."""
        return (
            f"{self.__class__.__name__}(" f"key='{self.key}', " f"operator='{self.operator}', " f"values={self.values})"
        )

    def __hash__(self) -> int:
        """Return a hash reprensetation of this object."""
        return hash(self.__repr__())


class ConditionCollection(Dict[ConditionKey, Dict[ConditionOperator, List[ConditionValue]]]):
    """A representation of a statement condition."""

    @property
    def conditions(self) -> List[Condition]:
        """Return a list of Condition Shards."""
        return Condition.factory(self)

    def __eq__(self, other: object) -> bool:
        """Determine whether this object and another object are equal.

        Parameters:
            other: The object to compare this one to.

        Raises:
            ValueError: When the object we are compared with is not of the same type.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        return self.conditions == other.conditions



"""Parent class for EffectiveAction, EffectiveResource, EffectivePrincipal."""
from typing import Any, Callable, Dict, FrozenSet, Generic, Iterator, List, Optional, Type, TypeVar, Union

from .protocols import ARPProtocol

T = TypeVar("T", bound=ARPProtocol)


class EffectiveARP(Generic[T]):
    """EffectiveARPs are the representation of the difference between an ARP and its exclusion.

    The allowed actions is the difference (subtraction) of the excluded ARPs
    from the included ARP.
    """

    #: Inclusion must be a superset of any exclusions
    inclusion: T

    #: Exclusions must always be a subset of the include and must not be subsets of each other
    exclusions: FrozenSet[T]

    #: The type of ARP we're subclassed for.
    _arp_type: Type[T]

    def __init__(self, inclusion: T, exclusions: Optional[FrozenSet[T]] = None) -> None:
        self.inclusion = inclusion
        self.exclusions = exclusions or frozenset()
        if not all([isinstance(action, self._arp_type) for action in [self.inclusion, *self.exclusions]]):
            raise ValueError(f"All inclusions and exclusions must be type {self._arp_type.__name__}")

    def union(self, other: object) -> List["EffectiveARP[T]"]:
        """Combine this object with another object of the same type.

        Parameters:
            other: The object to combine with this one.

        Raises:
            ValueError: If ``other`` is not the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot union {self.__class__.__name__} with {other.__class__.__name__}")
        if self.inclusion.issubset(other.inclusion) and not other.in_exclusions(self.inclusion):
            return [other]
        if other.inclusion.issubset(self.inclusion) and not self.in_exclusions(other.inclusion):
            return [self]
        return [self, other]

    def difference(self, other: object) -> List["EffectiveARP[T]"]:
        """Calculate the difference between this and another object of the same type.

        Effectively subtracts the inclusions of ``other`` from ``self``.
        This is useful when applying denies (``other``) to allows (``self``).

        Parameters:
            other: The object to subtract from this one.

        Raises:
            ValueError: If ``other`` is not the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot diff {self.__class__.__name__} with {other.__class__.__name__}")
        if self.inclusion.issubset(other.inclusion):
            return []
        if not other.inclusion.issubset(self.inclusion):
            return [self]
        if self.in_exclusions(other.inclusion):
            return [self]
        if not other.exclusions:
            # Just add the other's inclusion to self's exclusions
            return [self.__class__(self.inclusion, frozenset(set(self.exclusions).union(set({other.inclusion}))))]
        # If the other has its own exclusions we need to represent this as two seperate items
        # We need to add the inclusion from other to the exclusions of self and create new items for
        # each exclusion of other.
        # See docs for more details.
        new_self = self.__class__(self.inclusion, frozenset(set(self.exclusions).union(set({other.inclusion}))))
        new_others = [self.__class__(other_exclusion) for other_exclusion in other.exclusions]
        return [new_self, *new_others]

    def intersection(self, other: object) -> Optional["EffectiveARP[T]"]:
        """Calculate the intersection between this object and another object of the same type.

        Parameters:
            other: The object to intersect with this one.

        Raises:
            ValueError: if ``other`` is not hte same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot intersect {self.__class__.__name__} with {other.__class__.__name__}")

        if not self.inclusion.issubset(other.inclusion) and not other.inclusion.issubset(self.inclusion):
            return None
        if self.in_exclusions(other.inclusion):
            return None
        if self.inclusion.issubset(other.inclusion):
            if not other.exclusions:
                return self
            self_with_others_exclusions_added = self.__class__(
                self.inclusion, frozenset(set(self.exclusions).union(set(other.exclusions)))
            )
            return self_with_others_exclusions_added
        other_with_self_exclusions_added = self.__class__(
            other.inclusion, frozenset(set(self.exclusions.union(set(other.exclusions))))
        )

        return other_with_self_exclusions_added

    def issubset(self, other: object) -> bool:
        """Whether this object contains all the elements of another object (i.e. is a subset of the other object).

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        if not self.inclusion.issubset(other.inclusion):
            return False
        if self.in_exclusions(other.inclusion) or any(
            self_exclusion.issubset(other_exclusion)
            for self_exclusion in self.exclusions
            for other_exclusion in other.exclusions
        ):
            return False
        return True

    def in_exclusions(self, other: T) -> bool:
        """Check if the ARP is contained within or equal to any of the exclusions.

        Parameters:
            other: The object to look for in the exclusions of this object.
        """
        return any(other.issubset(exclusion) for exclusion in self.exclusions)

    def __eq__(self, other: object) -> bool:
        """Whether this object contains (but is not equal to) another object.

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        return self.inclusion == other.inclusion and self.exclusions == other.exclusions

    def __repr__(self) -> str:
        """Return an insantiable representation of this object."""
        return f"{self.__class__.__name__}(inclusion={repr(self.inclusion)}, exclusions={self.exclusions})"

    def dict(self, *args, **kwargs) -> Dict[str, Any]:
        """Return a dictionary representation of this object.

        Parameters:
            *args: Arguments to Pydantic dict method.
            **kwargs: Arguments to Pydantic dict method.

        """
        result: Dict[str, Union[T, FrozenSet[T]]] = {"inclusion": self.inclusion}
        if not kwargs.get("exclude_defaults") or self.exclusions != frozenset():
            result.update({"exclusions": self.exclusions})
        return result

    @classmethod
    def __get_validators__(cls) -> Iterator[Callable]:
        """Return noop validator as we don't particularly need validation.

        We just need Pydantic to accept this as a valid type to populate in PolicyShard.
        """
        yield lambda x: x



"""Action class."""
from fnmatch import fnmatch

from .effective_arp import EffectiveARP
from .models import CaseInsensitiveString


class Action(CaseInsensitiveString):
    """Actions are case insensitive.

    .. epigraph::

        "The prefix and the action name are case insensitive"

        -- `IAM JSON policy elements: Action
        <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_action.html>`__
    """

    def issubset(self, other: object) -> bool:
        """Whether this object contains all the elements of another object (i.e. is a subset of the other object).

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        return fnmatch(self.lower(), other.lower())

    def __lt__(self, other: object) -> bool:
        """Whether this object contains but is not equal to (i.e. a proper subset) another object.

        Parameters:
            other: The object to determine if our object contains (but is not equal to).

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        if self == other:
            return False
        return self.issubset(other)

    def __contains__(self, other: object) -> bool:
        """Not Implemented.

        Parameters:
            other: The object to see if this object contains.

        Raises:
            NotImplementedError: This method is not implemented.
        """
        raise NotImplementedError()


class EffectiveAction(EffectiveARP[Action]):
    """EffectiveActions are the representation of the difference between an Action and its exclusion.

    The allowed actions is the difference (subtraction) of the excluded Actions
    from the included action.
    """

    _arp_type = Action



"""Generic Models."""


class CaseInsensitiveString(str):
    """A case insensitive string to aid comparison."""

    def __eq__(self, other: object) -> bool:
        """Determine whether this object and another object are equal.

        Parameters:
            other: The object to compare this one to.

        Raises:
            ValueError: When the object we are compared with is not of the same type.
        """
        if not isinstance(other, (self.__class__, str)):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        return self.lower() == other.lower()

    def __hash__(self) -> int:
        """Compute the hash for this object."""
        return super().__hash__()

    def __repr__(self) -> str:
        """Return an instantiable representation of this object."""
        return f"{self.__class__.__name__}('{self}')"



"""Core Policy class."""
from typing import List

from pydantic import BaseModel

from .policy_shard import PolicyShard
from .statement import Statement
from .utils import to_pascal


class Policy(BaseModel):
    """Main policy class.

    Example:
        Create a policy from a dictionary.

            >>> from policyglass import Policy
            >>> Policy(**{
            ...     "Version": "2012-10-17",
            ...     "Statement": [
            ...         {
            ...             "Effect": "Allow",
            ...             "Action": [
            ...                 "s3:*"
            ...             ],
            ...             "Resource": "*"
            ...         }
            ...     ]
            ... })
            Policy(version='2012-10-17',
                statement=[Statement(effect='Allow',
                    action=[Action('s3:*')],
                    not_action=None,
                    resource=[Resource('*')],
                    not_resource=None, principal=None,
                    not_principal=None,
                    condition=None)])
    """

    version: str
    statement: List[Statement]

    class Config:
        """Configure the pydantic BaseModel."""

        alias_generator = to_pascal

    @property
    def policy_shards(self) -> List[PolicyShard]:
        """Shatter this policy into a number :class:`policyglass.policy_shard` objects."""
        result = []
        for statement in self.statement:
            result.extend(statement.policy_shards)
        return result

    def policy_json(self) -> str:
        """Return a valid policy JSON from this policy."""
        return self.json(by_alias=True, exclude_none=True)



"""Protocol types used by PolicyGlass."""
import sys

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


class ARPProtocol(Protocol):
    """Protocol which Actions, Resources, and Principals must implement."""

    def __eq__(self, other: object) -> bool:
        """Determine whether this object and another object are equal.

        Parameters:
            other: The object to compare this one to.
        """
        ...

    def issubset(self, other: object) -> bool:
        """Whether this object contains all the elements of another object (i.e. is a subset of the other object).

        Parameters:
            other: The object to determine if our object contains.
        """
        ...

    def __lt__(self, other: object) -> bool:
        """Whether this object contains but is not equal to (i.e. a proper subset) another object.

        Parameters:
            other: The object to determine if our object contains (but is not equal to).
        """
        ...



"""PolicyShards are a simplified representation of policies."""

import json
from typing import Any, Dict, FrozenSet, List, Optional

from pydantic import BaseModel

from .action import Action
from .condition import Condition
from .effective_arp import EffectiveARP
from .principal import Principal
from .resource import Resource


def dedupe_policy_shards(shards: List["PolicyShard"]) -> List["PolicyShard"]:
    """Dedupe policy shards that are subsets of each other.

    Parameters:
        shards: The shards to deduplicate.
    """
    deduped_shards: List[PolicyShard] = []
    for undeduped_shard in shards:
        if any(undeduped_shard.issubset(deduped_shard) for deduped_shard in deduped_shards):
            continue

        deduped_shards.append(undeduped_shard)

    return deduped_shards


def policy_shards_effect(shards: List["PolicyShard"]) -> List["PolicyShard"]:
    """Calculate the effect of merging allow and deny shards together.

    Parameters:
        shards: The shards to caclulate the effect of.
    """
    allow_shards = [shard for shard in shards if shard.effect == "Allow"]
    deny_shards = [shard for shard in shards if shard.effect == "Deny"]

    # This code is ugly because DIFFERENCE takes in a single shard and yields a list of them.
    merged_allow_shards = []
    for allow_shard in allow_shards:
        allow_candidates = [allow_shard]
        for deny_shard in deny_shards:
            result = []
            for allow_candidate in allow_candidates:
                result.extend(allow_candidate.difference(deny_shard))
            allow_candidates = result
        if allow_candidates:
            merged_allow_shards.extend(allow_candidates)
    return merged_allow_shards


def policy_shards_to_json(shards: List["PolicyShard"], exclude_defaults=False, **kwargs) -> str:
    """Convert a list of :class:`awspolicy.policy_shard.PolicyShard` objects to JSON.

    Parameters:
        shards: The list of shards to convert.
        exclude_defaults: Whether to exclude default values (e.g. empty lists) from the output.
        **kwargs: keyword arguments passed on to :func:`json.dumps`
    """
    return json.dumps([json.loads(shard.json(exclude_defaults=exclude_defaults)) for shard in shards], **kwargs)


class PolicyShard(BaseModel):
    """A PolicyShard is part of a policy broken down in such a way that it can be deduplicated and collapsed."""

    effect: str
    effective_action: EffectiveARP[Action]
    effective_resource: EffectiveARP[Resource]
    effective_principal: EffectiveARP[Principal]
    conditions: FrozenSet[Condition]
    not_conditions: FrozenSet[Condition]

    def __init__(
        self,
        effect: str,
        effective_action: EffectiveARP[Action],
        effective_resource: EffectiveARP[Resource],
        effective_principal: EffectiveARP[Principal],
        conditions: Optional[FrozenSet[Condition]] = None,
        not_conditions: Optional[FrozenSet[Condition]] = None,
    ) -> None:
        super().__init__(
            effect=effect,
            effective_action=effective_action,
            effective_resource=effective_resource,
            effective_principal=effective_principal,
            conditions=conditions or frozenset(),
            not_conditions=not_conditions or frozenset(),
        )

    def union(self, other: object) -> List["PolicyShard"]:
        """Combine this object with another object of the same type.

        Parameters:
            other: The object to combine with this one.

        Raises:
            ValueError: If ``other`` is not the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot union {self.__class__.__name__} with {other.__class__.__name__}")
        if not self.conditions == other.conditions:
            return [self, other]

        return [
            self.__class__(
                effect=self.effect,
                effective_action=effective_action,
                effective_resource=effective_resource,
                effective_principal=effective_principal,
                conditions=self.conditions,
            )
            for effective_action in self.effective_action.union(other.effective_action)
            for effective_resource in self.effective_resource.union(other.effective_resource)
            for effective_principal in self.effective_principal.union(other.effective_principal)
        ]

    def difference(self, other: object) -> List["PolicyShard"]:
        """Calculate the difference between this and another object of the same type.

        Effectively subtracts the inclusions of ``other`` from ``self``.
        This is useful when applying denies (``other``) to allows (``self``).

        Parameters:
            other: The object to subtract from this one.

        Raises:
            ValueError: If ``other`` is not the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot diff {self.__class__.__name__} with {other.__class__.__name__}")

        intersection_action = self.effective_action.intersection(other.effective_action)
        intersection_resource = self.effective_resource.intersection(other.effective_resource)
        intersection_principal = self.effective_principal.intersection(other.effective_principal)

        if not intersection_action or not intersection_resource or not intersection_principal:
            # Shards do not overlap
            return [self]

        effective_actions = self.effective_action.difference(other.effective_action)
        effective_resources = self.effective_resource.difference(other.effective_resource)
        effective_principals = self.effective_principal.difference(other.effective_principal)

        if not effective_actions and not effective_resources and not effective_principals:
            # Shards overlap wholly
            return []

        result = [
            self.__class__(
                effect=self.effect,
                effective_action=effective_action,
                effective_resource=effective_resource,
                effective_principal=effective_principal,
                conditions=self.conditions,
                not_conditions=self.not_conditions,
            )
            for effective_action in effective_actions or [self.effective_action]
            for effective_resource in effective_resources or [self.effective_resource]
            for effective_principal in effective_principals or [self.effective_principal]
        ]

        if self.conditions != other.conditions:
            # If the conditions differ return the intersection of the two shards back into the result.
            # This results in an uncomplicated difference (already in the result) with a conditional intersection.
            result.append(
                self.__class__(
                    effect=self.effect,
                    effective_action=intersection_action,
                    effective_resource=intersection_resource,
                    effective_principal=intersection_principal,
                    conditions=self.conditions,
                    not_conditions=other.conditions,
                )
            )
        return result

    def issubset(self, other: object) -> bool:
        """Whether this object contains all the elements of another object (i.e. is a subset of the other object).

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")

        return (
            self.effective_action.issubset(other.effective_action)
            and self.effective_resource.issubset(other.effective_resource)
            and self.effective_principal.issubset(other.effective_principal)
            and self.conditions == other.conditions
            and self.not_conditions == other.not_conditions
            and self.effect == other.effect
        )

    def dict(self, *args, **kwargs) -> Dict[str, Any]:
        """Convert instance to dict representation of it.

        Parameters:
            *args: Arguments to Pydantic dict method.
            **kwargs: Arguments to Pydantic dict method.

        Overridden from BaseModel so that when converting conditions to dict they don't suffer from being unhashable
        when placed in a set.
        """
        result = {}
        for attribute_name, attribute_value in self:
            if hasattr(attribute_value, "dict"):
                result[attribute_name] = attribute_value.dict(*args, **kwargs)
            elif isinstance(attribute_value, (set, frozenset)):
                value = list(attribute_value)
                if not kwargs.get("exclude_defaults") or value != []:
                    result[attribute_name] = value

        return result

    def __repr__(self) -> str:
        """Return an instantiable representation of this object."""
        return (
            f"{self.__class__.__name__}(effect='{self.effect}', "
            f"effective_action={self.effective_action}, "
            f"effective_resource={self.effective_resource}, "
            f"effective_principal={self.effective_principal}, "
            f"conditions={self.conditions}, "
            f"not_conditions={self.not_conditions})"
        )

    def __str__(self) -> str:
        """Return a stringy representation of this object."""
        return repr(self)

    def __eq__(self, other: object) -> bool:
        """Determine whether this object and another object are equal.

        Parameters:
            other: The object to compare this one to.

        Raises:
            ValueError: When the object we are compared with is not of the same type.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        return (
            self.effective_action == other.effective_action
            and self.effective_resource == other.effective_resource
            and self.effective_principal == other.effective_principal
            and self.conditions == other.conditions
        )
